#include "pipe_pager.h"
#include "pipe_object_server.h"

/* Initialise a pager for a pipe with a shared page mapper for moderating
   access to loaded pages. At first, no mapper will be configured: this must be
   done by requesting a region. */

PipePager::PipePager(PagesConserving *pages, PipePaging *paging, bool writer)
: Pager(NULL, writer ? L4_FPAGE_RW : L4_FPAGE_RO),
  _pages(pages), _paging(paging), _writer(writer)
{
    /* Set the size of each paged region to the amount allocation for such
       regions. */

    _size = _pages->allocation();
}

int PipePager::expected_items()
{
    return PipeObject_expected_items;
}

ipc_server_handler_type PipePager::handler()
{
    return (ipc_server_handler_type) handle_PipeObject;
}



/* Support paging. */

long PipePager::map(unsigned long offset, l4_addr_t hot_spot, flags_t flags, l4_snd_fpage_t *region)
{
    return Pager::map(offset, hot_spot, flags, region);
}



/* Return details of the current region. */

long PipePager::current_region(offset_t *populated_size, offset_t *size)
{
    if (_mapper != NULL)
    {
        *populated_size = _mapper->get_data_size();
        *size = _size;
        return L4_EOK;
    }
    else
        return -L4_EIO;
}

/* Obtain the next region and its details. */

long PipePager::next_region(offset_t *populated_size, offset_t *size)
{
    /* Obtain a new region if writing. */

    if (_writer)
        return next_region_for_writer(populated_size, size);
    else
        return next_region_for_reader(populated_size, size);
}

long PipePager::next_region_for_reader(offset_t *populated_size, offset_t *size)
{
    /* Obtain the next region if a current region is defined. */

    if (_mapper != NULL)
    {
        PageMapper *mapper = _paging->next_region();

        /* Return an error if no next region currently exists. */

        if (mapper == NULL)
            return -L4_EIO;

        /* Detach and discard the current page mapper. */

        _mapper->detach();
        delete _mapper;

        _mapper = mapper;
    }

    /* Obtain the first region if no mapper is defined. */

    else
        _mapper = _paging->first_region();

    /* Return the details of the now-current region. */

    return current_region(populated_size, size);
}

long PipePager::next_region_for_writer(offset_t *populated_size, offset_t *size)
{
    /* Set the populated size before moving on. */

    if (_mapper != NULL)
        _mapper->set_data_size(*populated_size);

    /* A guaranteed amount of memory is reserved for the pipe. */

    if (!_pages->increase())
        return -L4_ENOMEM;

    /* Set the page mapper for the region. */

    _mapper = new PageMapper(&_accessor, _pages);
    _mapper->attach();

    /* Record the mapper as the latest region in the accessor. */

    _paging->add_region(_mapper);

    /* Set the region size to the allocated amount. */

    *size = _size;
    *populated_size = 0;

    return L4_EOK;
}

// vim: tabstop=4 expandtab shiftwidth=4
